package fr.smallcrew.rest.converter;

import org.reflections.Reflections;
import org.springframework.http.HttpInputMessage;
import org.springframework.http.HttpOutputMessage;
import org.springframework.http.MediaType;
import org.springframework.http.converter.AbstractHttpMessageConverter;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.http.converter.HttpMessageNotWritableException;

import javax.xml.bind.JAXBContext;
import javax.xml.bind.JAXBException;
import javax.xml.bind.Marshaller;
import javax.xml.bind.Unmarshaller;
import javax.xml.bind.annotation.XmlRootElement;
import java.io.IOException;
import java.util.Set;

import static com.google.common.base.Throwables.propagate;
import static com.google.common.collect.Sets.newHashSet;

public class JaxbHttpMessageConverter extends AbstractHttpMessageConverter<Object> implements RestConverter {
  public static final String DEFAULT_PACKAGE = "fr.smallcrew.rest";
  public static final String ENCODING = "UTF-8";

  private final Set<Class<?>> classes;
  private Marshaller jaxbMarshaller;
  private Unmarshaller jaxbUnmarshaller;

  public JaxbHttpMessageConverter() {
    super(MediaType.APPLICATION_XML);
    classes = newHashSet();
    addPackage(newHashSet(DEFAULT_PACKAGE));
    consolidate();
  }

  @Override
  public boolean canRead(Class<?> clazz, MediaType mediaType) {
    return canRead(mediaType);
  }

  @Override
  public boolean canWrite(Class<?> clazz, MediaType mediaType) {
    return canWrite(mediaType);
  }

  @Override
  public boolean supports(Class<?> clazz) {
    return true;
  }

  @Override
  @SuppressWarnings("all")
  public Object readInternal(Class<? extends Object> clazz, HttpInputMessage inputMessage) throws IOException, HttpMessageNotReadableException {
    try {
      return jaxbUnmarshaller.unmarshal(inputMessage.getBody());
    } catch (JAXBException e) {
      throw propagate(e);
    }
  }

  @Override
  public void writeInternal(Object o, HttpOutputMessage outputMessage) throws IOException, HttpMessageNotWritableException {
    try {
      jaxbMarshaller.marshal(o, outputMessage.getBody());
    } catch (JAXBException e) {
      throw propagate(e);
    }
  }

  @Override
  public void addPackage(Set<String> packageName) {
    for (String name : packageName) {
      for (Class<?> clazz : new Reflections(name).getTypesAnnotatedWith(XmlRootElement.class)) {
        classes.add(clazz);
      }
    }
  }

  @Override
  public void addClasses(Set<String> classNames) {
    try {
      for (String className : classNames) {
        classes.add(Class.forName(className));
      }
    } catch (Exception e) {
      throw propagate(e);
    }
  }

  @Override
  public void consolidate() {
    try {
      JAXBContext jaxbContext = JAXBContext.newInstance(classes.toArray(new Class[classes.size()]));
      jaxbMarshaller = jaxbContext.createMarshaller();
      jaxbMarshaller.setProperty(Marshaller.JAXB_FORMATTED_OUTPUT, true);
      jaxbMarshaller.setProperty(javax.xml.bind.Marshaller.JAXB_ENCODING, ENCODING);
      jaxbUnmarshaller = jaxbContext.createUnmarshaller();
    } catch (Exception e) {
      throw propagate(e);
    }
  }
}

