package fr.smallcrew.rest.exception;

import com.google.common.base.Throwables;
import fr.smallcrew.rest.data.ResourceException;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;
import org.springframework.beans.ConversionNotSupportedException;
import org.springframework.beans.TypeMismatchException;
import org.springframework.http.HttpStatus;
import org.springframework.http.converter.HttpMessageNotReadableException;
import org.springframework.http.converter.HttpMessageNotWritableException;
import org.springframework.web.HttpMediaTypeNotAcceptableException;
import org.springframework.web.HttpMediaTypeNotSupportedException;
import org.springframework.web.HttpRequestMethodNotSupportedException;
import org.springframework.web.bind.MissingServletRequestParameterException;
import org.springframework.web.servlet.ModelAndView;
import org.springframework.web.servlet.View;
import org.springframework.web.servlet.handler.AbstractHandlerExceptionResolver;
import org.springframework.web.servlet.mvc.multiaction.NoSuchRequestHandlingMethodException;
import org.springframework.web.servlet.view.ContentNegotiatingViewResolver;

import javax.servlet.http.HttpServletRequest;
import javax.servlet.http.HttpServletResponse;
import java.util.Map;

import static com.google.common.base.Throwables.propagate;
import static com.google.common.collect.Maps.newHashMap;
import static org.springframework.http.HttpStatus.BAD_REQUEST;
import static org.springframework.http.HttpStatus.INTERNAL_SERVER_ERROR;
import static org.springframework.http.HttpStatus.METHOD_NOT_ALLOWED;
import static org.springframework.http.HttpStatus.NOT_ACCEPTABLE;
import static org.springframework.http.HttpStatus.NOT_FOUND;
import static org.springframework.http.HttpStatus.UNSUPPORTED_MEDIA_TYPE;

public class ResourceExceptionHandler extends AbstractHandlerExceptionResolver {
  public static final String EXCEPTION_HAPPENED = ResourceExceptionHandler.class.getSimpleName() + "$EXCEPTION_HAPPENED";
  public static final HttpStatus DEFAULT_ERROR = INTERNAL_SERVER_ERROR;
  public static final String MODEL_NAME = "exception";

  private final Logger logger = LoggerFactory.getLogger(ResourceExceptionHandler.class);
  private final Map<Class<?>, HttpStatus> exceptionDefinitions;
  private final ContentNegotiatingViewResolver viewResolver;

  public ResourceExceptionHandler(ContentNegotiatingViewResolver viewResolver) {
    this.viewResolver = viewResolver;
    this.exceptionDefinitions = newHashMap();
    addDefinition(ConversionNotSupportedException.class, INTERNAL_SERVER_ERROR);
    addDefinition(HttpMediaTypeNotAcceptableException.class, NOT_ACCEPTABLE);
    addDefinition(HttpMediaTypeNotSupportedException.class, UNSUPPORTED_MEDIA_TYPE);
    addDefinition(HttpMessageNotReadableException.class, BAD_REQUEST);
    addDefinition(HttpMessageNotWritableException.class, INTERNAL_SERVER_ERROR);
    addDefinition(HttpRequestMethodNotSupportedException.class, METHOD_NOT_ALLOWED);
    addDefinition(MissingServletRequestParameterException.class, BAD_REQUEST);
    addDefinition(NoSuchRequestHandlingMethodException.class, NOT_FOUND);
    addDefinition(TypeMismatchException.class, BAD_REQUEST);
    addDefinition(ResourceNotFoundException.class, NOT_FOUND);
    initialize();
  }

  public void addDefinition(Class<? extends Throwable> throwable, HttpStatus status) {
    exceptionDefinitions.put(throwable, status);
  }
  
  public void initialize() {
    setMappedHandlerClasses(exceptionDefinitions.keySet().toArray(new Class[exceptionDefinitions.size()]));
  }

  @Override
  public ModelAndView resolveException(HttpServletRequest request, HttpServletResponse response, Object handler,
                                       Exception exception) {
    return doResolveException(request, response, handler, exception);
  }

  @Override
  protected ModelAndView doResolveException(HttpServletRequest request, HttpServletResponse response, Object handler,
                                            Exception exception) {
    try {
      logger.warn("Rest exception", exception);
      View view = viewResolver.resolveViewName(null, request.getLocale());
      return new ModelAndView(view, MODEL_NAME, newResourceException(request, response, exception));
    } catch (Exception e) {
      logger.error("Can't resolve exception", e);
      throw propagate(e);
    }
  }

  private ResourceException newResourceException(HttpServletRequest request, HttpServletResponse response, Exception exception) {
    Throwable throwable = Throwables.getRootCause(exception);
    HttpStatus status = httpStatus(throwable);
    response.setStatus(status.value());
    request.setAttribute(EXCEPTION_HAPPENED, true);
    return new ResourceException(status, throwable);
  }

  private HttpStatus httpStatus(Throwable exception) {
    HttpStatus httpStatus = exceptionDefinitions.get(exception.getClass());
    if (httpStatus == null) {
      httpStatus = DEFAULT_ERROR;
    }
    return httpStatus;
  }
}
