package fr.smallcrew.rest.servlet.filter;

import com.google.common.annotations.VisibleForTesting;
import org.slf4j.Logger;
import org.slf4j.LoggerFactory;

import javax.servlet.Filter;
import javax.servlet.FilterChain;
import javax.servlet.FilterConfig;
import javax.servlet.ServletException;
import javax.servlet.ServletRequest;
import javax.servlet.ServletResponse;
import javax.servlet.annotation.WebFilter;
import javax.servlet.http.HttpServletRequest;
import java.io.IOException;

@WebFilter(urlPatterns = "/api/*")
public class ExecutionTimeFilter implements Filter {
  public static final String START_EXECUTION_TIME = ExecutionTimeFilter.class.getSimpleName() + "START_EXECUTION_TIME";
  public static final String STOP_EXECUTION_TIME = ExecutionTimeFilter.class.getSimpleName() + "STOP_EXECUTION_TIME";

  private final Logger logger;

  public ExecutionTimeFilter() {
    logger = LoggerFactory.getLogger(ExecutionTimeFilter.class);
  }

  @VisibleForTesting
  ExecutionTimeFilter(Logger logger) {
    this.logger = logger;
  }

  @Override
  public void init(FilterConfig filterConfig) throws ServletException {
  }

  @Override
  public void doFilter(ServletRequest request, ServletResponse response, FilterChain chain) throws IOException, ServletException {
    HttpServletRequest servletRequest = (HttpServletRequest) request;
    String logMessage = newLogMessage(servletRequest);

    try {
      markStartExecution(request, logMessage);
      chain.doFilter(request, response);
    } finally {
      markExecutionStop(request, logMessage);
    }
  }

  @Override
  public void destroy() {
  }

  private void markStartExecution(ServletRequest request, String logMessage) {
    request.setAttribute(START_EXECUTION_TIME, System.currentTimeMillis());
    logger.debug("begin => {}", logMessage);
  }

  private void markExecutionStop(ServletRequest request, String logMessage) {
    Long begin = (Long) request.getAttribute(START_EXECUTION_TIME);
    long end = System.currentTimeMillis() - begin;
    request.setAttribute(STOP_EXECUTION_TIME, end);
    logger.debug("stop => {} executed in {}ms", logMessage, end);
  }

  private String newLogMessage(HttpServletRequest request) {
    if (logger.isDebugEnabled()) {
      StringBuilder log = new StringBuilder();
      if (request.getUserPrincipal() != null) {
        log.append(request.getUserPrincipal());
      } else {
        log.append(request.getRemoteHost());
      }
      log.append("@");
      log.append(request.getLocalAddr()).append(":");
      log.append(request.getLocalPort());
      log.append(request.getRequestURI());
      if (request.getQueryString() != null) {
        log.append("?").append(request.getQueryString());
      }
      return log.toString();
    }
    return null;
  }
}
